/*** news.c    - print system news	(C) 1993/1994
****
**** intruder@link.hacktic.nl
***/

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>
#include <time.h>
#include <fcntl.h>
#include <pwd.h>

#define VERSION 	"v0.5"
#define PAGER		"more"
#define ERROR		strerror(errno)
#define DATEFORMAT	"(%b %d %Y)"

char  *myname;
int   verbose=0;

FILE  *open_pager();
int   cat(char *);
int   more(FILE *,char *);
int   fcat(int, int);
void  print_usage();
void  doexpire(int, char *);

/**
*** MAIN
**/
void main(int argc, char **argv)
{
  DIR	 *dfd;
  FILE	 *pfd=(FILE *)NULL;
  struct dirent *dir;
  struct stat st;
  struct passwd *pw;
  time_t nt=0;
  char	 *home, line[1024];
  char	 tmp[512];
  char	 newsnames[80];
  char	 *excludelist=0;
  int	 f_all=0;
  int	 f_items=0;
  int	 f_names=0;
  int	 f_pager=0;
  int	 f_expire=0;
  int	 f_oneperline=0;
  int	 f_datestamp=0;
  int	 ch, icount=0;
  int	 expireover=0;
  int	 avc;
  extern int  optind;
  extern char *optarg;

  myname = (char *)argv[0];

  while ((ch = getopt(argc, argv, "ade:snlpvx:?h")) != EOF)
    switch(ch)
     {
	   case 'a':            /* show all articles    */
		f_all++;
		break;
	   case 'd':            /* show date of article */
		f_datestamp++;
		break;
	   case 'e':            /* expire articles      */
		expireover = atoi(optarg);
		f_expire++;
		break;
	   case 's':
		f_items++;	/* show # of articles	*/
		break;
	   case 'n':
		f_names++;	/* show names of articles */
		break;
	   case 'l':
		f_oneperline++; /* show one article per line */
		break;
	   case 'p':            /* pipe through $PAGER  */
		f_pager++;
		break;
	   case 'v':
		verbose++;
		break;
	   case 'x':
		excludelist = optarg;
		break;
	   case '?':            /* hm. what could this be? */
	   case 'h':
	   default:
		print_usage();
		exit(0);
		break;
     }

    if(f_expire)
     doexpire(expireover, excludelist);

  home = getenv("HOME");
  if(!home)
    exit(0);

  sprintf(line, "%s/%s", home, NEWSTIME);
  if(stat(line, &st) == -1)
    {
      if(errno != ENOENT)
	exit(errno);
      if(!f_all && !f_items && !f_names && !argv[optind])
       close(creat(line, 0600));
    }
  else
   nt = st.st_mtime;

 if(!f_all && !f_items && !f_names && !argv[optind])
  if(utime(line, (struct utimbuf *)0) == -1)
   {
     fprintf(stderr, "%s: (%s) %s\n", myname, line, ERROR);
   }

  dfd = opendir(NEWSDIR);
  if(!dfd)
   {
     fprintf(stderr, "%s: opendir(%s): %s\n", myname, NEWSDIR, ERROR);
     exit(errno);
   }

  while((dir=readdir(dfd)))
   {
     if(!strcmp(dir->d_name, ".") || !strcmp(dir->d_name, ".."))
       continue;
     if(argv[optind])
      {
	avc=optind;
	for(avc=optind;avc<argc;avc++)
	 {
	   if(strstr(dir->d_name, argv[avc]))
	    break;
	 }
	if(!argv[avc])
	 continue;
      }
     sprintf(line, "%s/%s", NEWSDIR, dir->d_name);
     if(stat(line, &st) == -1)
      {
	fprintf(stderr, "%s: (%s) %s\n", myname, line, ERROR);
	continue;
      }
     if(nt < st.st_mtime || f_all || argv[optind])
      {
	if(!f_names && !f_items)
	 {
	    if(f_pager && (!pfd))
	     {
	       pfd=open_pager();
	       if(!pfd)
		exit(1);
	     }
	    pw = getpwuid(st.st_uid);
	    fprintf(pfd?pfd:stdout, "\n** %s (%s)  %s\n",dir->d_name,
	     pw ? pw->pw_name : "unknown", ctime(&st.st_mtime));
	    fflush(pfd?pfd:stdout);
	    if(f_pager)
	     {
	       more(pfd, line);
	     }
	    else
	     cat(line);
	 }
	if(f_names)
	  {
	    if(f_oneperline)
	     {
	       if(f_datestamp)
		{
		  strftime(tmp, 512, DATEFORMAT, gmtime(&st.st_mtime));
		  printf("NEWS: %s %s\n", tmp, dir->d_name);
		}
	       else
		{
		  printf("NEWS: %s\n", dir->d_name);
		}
	     }
	    else
	     {
	       if((strlen(dir->d_name) +
		   strlen(newsnames) +
		   strlen(myname) + 1 + 2) > 80)
		{
		   printf("NEWS: %s\n", newsnames);
		   strcpy(newsnames, "");
		}
	       strcat(newsnames, dir->d_name);
	       strcat(newsnames, " ");
	     }
	  }
	icount++;
      }
   }
  if(f_names && icount && newsnames[0])
     printf("NEWS: %s\n", newsnames);
  if(f_items)
     printf("NEWS: %d news article%s\n", icount, icount==1 ? "" : "s");
  closedir(dfd);
  if(pfd)
   pclose(pfd);
 exit(0);
}

/**
*** print a file to stdout
**/
int cat(char *file)
{
  FILE *fd;

  fd = fopen(file, "r");
  if(!fd)
   {
     fprintf(stderr, "%s: (%s) %s\n", myname, file, ERROR);
     return 0;
   }

  fcat(fileno(fd), fileno(stdout));
 fclose(fd);
 return 1;
}

/**
*** pipe a file through $PAGER
**/
int more(FILE *pfd, char *file)
{
  FILE *fd;

  fd = fopen(file, "r");
  if(!fd)
    {
       fprintf(stderr, "%s: (%s) %s\n", myname, file, ERROR);
       return 0;
     }
  fcat(fileno(fd), fileno(pfd));
  fclose(fd);
 return 1;
}

/**
*** setup a pipe to $PAGER
**/
FILE *open_pager()
{
  FILE *pfd;

   pfd = popen(getenv("PAGER") ? getenv("PAGER") : PAGER, "w");
   if(!pfd)
     {
       fprintf(stderr, "%s: popen() failed: %s\n", myname, ERROR);
       return (FILE *)0;
     }
 return pfd;
}

int fcat(int fdin, int fdout)
{
  int i;
  char buf[1024];

  while((i=read(fdin, buf, 1024)))
   {
     if(i == -1)
       {
	  fprintf(stderr, "%s: %s\n", myname, ERROR);
	  return 0;
       }
     write(fdout, buf, i);
   }
  return 1;
}


/**
*** expire old news
**/
void doexpire(int expireover, char *xlist)
{
  struct stat st;
  struct dirent *dir;
  DIR	 *dfd;
  time_t t, tt;

  dfd = opendir(NEWSDIR);
  if(!dfd)
   {
     fprintf(stderr, "Error opening NEWSDIR (%s): %s\n", NEWSDIR, ERROR);
     exit(ENOENT);
   }

  chdir(NEWSDIR);
  time(&t);
  tt = expireover * (60 * 60 * 24);

  while((dir=readdir(dfd)))
   {
     if(!strcmp(dir->d_name, ".") || !strcmp(dir->d_name, ".."))
      continue;
     if(strstr(xlist, dir->d_name))
      continue;
     stat(dir->d_name, &st);
     if(st.st_mtime < t - tt)
      if(remove(dir->d_name) == -1)
       {
	 fprintf(stderr, "Error: remove (%s): %s\n", dir->d_name, ERROR);
       }
      else
       if(verbose)
	printf("Expiring article: %s\n", dir->d_name);
   }
  closedir(dfd);
  exit(0);
}

void print_usage()
{
  fprintf(stderr, "news %s  - Display system news - (C)'93/'94 intruder@link.hacktic.nl\n", VERSION);
  fprintf(stderr, "Usage: news [-flags] [[article1] [article2] .. ]\n\n");
  fprintf(stderr, "        -a     Display all news\n");
  fprintf(stderr, "        -d     Add datestamp\n");
  fprintf(stderr, "        -n     Display news article names only\n");
  fprintf(stderr, "        -l     One article name per line\n");
  fprintf(stderr, "        -p     Pipe article through $PAGER\n");
  fprintf(stderr, "        -s     Display number of news articles\n");
  fprintf(stderr, "        -e #   Expire news older than # days\n");
  fprintf(stderr, "        -x a,b List of articles to exclude when expiring\n\n");
}
